
class Nbody:
  
  def print_transformations(self):
    print("I am transformations")

  def UU(self):
    """
    Return the gas specific energy of the model.
    The output is an nx1 float array.
    """
    return self.u

